// by Ido Gendel, 2014
// Windows only

unit COMPort;

{$mode objfpc}{$H+}

interface

uses
  SdpoSerial, classes;

ResourceString
  ErrMsgBaudrateChange       = 'Error: Baudrate change failed';
  ErrMsgRegistryAccess       = 'Error: Cannot access registry';
  ErrMsgResetDTR             = 'Error: Setting DTR off/on failed';
  ErrMsgCreatePortStringList = 'Error: Creating StringList for ports failed';
  ErrMsgReadPorts            = 'Error: Cannot read available ports';
  ErrMsgSetPort              = 'Error: Cannot set port';

const
  NO_PORT_STR = '(None)';

type
  TSerial = class
    private

      FSer            : TSdpoSerial;
      FBaudrate       : Int32;
      FPort           : string;
      FAvailablePorts : TStringList;

      procedure SetBaudRate(const Rate : Int32);
      function GetDataAvailable : Boolean;
      function GetAvailablePorts : Boolean;
      procedure SetPort(const s : string);

    public

      constructor Create; overload;

      procedure ResetDTR;
      procedure Send(const s : string);
      function Receive : string;
      procedure AssignRxHandler(p : TNotifyEvent);

      property BaudRate : Int32 read FBaudRate write SetBaudRate;
      property DataAvailable : Boolean read GetDataAvailable;
      property AvailablePorts : TStringList read FAvailablePorts;
      property AvailablePortsChanged : Boolean read GetAvailablePorts;
      property Port : string read FPort write SetPort;

  end;

var
  Serial : TSerial;

implementation
uses
  SysUtils, Registry, SMDUtils;

const
  DTRResetLengthMS = 2;

var
  RegAccess : TRegistry = nil;

procedure TSerial.SetBaudRate(const Rate : Int32);
begin

  FBaudrate := Rate;
  try

    // Had to hack the original SdpoSerial.SetBaudRate because it only accepted baudrates
    // from a constant list. Last two parameters are SoftwareFlow and HardwareFlow, which
    // I can't access, but under normal circumstances should be both False
    FSer.SynSer.Config(Rate, ConstsBits[FSer.DataBits], ConstsParity[FSer.Parity],
                       ConstsStopBits[FSer.StopBits], False, False);

  except
    raise Exception.Create(ErrMsgBaudrateChange);
  end;

end;

function TSerial.GetDataAvailable : Boolean;
begin
  Result := FSer.DataAvailable;
end;

// Returns True if there was a change (and on the first call)
function TSerial.GetAvailablePorts : Boolean;
var
  PortNames : TStringList;
  i         : Integer;

begin

  try
    if NOT assigned(FAvailablePorts) then FAvailablePorts := TStringList.Create;
    PortNames := TStringList.Create;
  except
    raise Exception.Create(ErrMsgCreatePortStringList);
  end;

  try

    // Get port name keys
    RegAccess.OpenKeyReadOnly('\HARDWARE\DEVICEMAP\SERIALCOMM\');
    RegAccess.GetValueNames(PortNames);
    // Get port name values by keys
    for i := 0 to PortNames.Count - 1 do
     PortNames[i] := RegAccess.ReadString(PortNames[i]);

  except
    raise Exception.Create(ErrMsgReadPorts);
  end;

  RegAccess.CloseKey;

  if PortNames.CommaText <> FAvailablePorts.CommaText then
   begin
     FAvailablePorts.CommaText := PortNames.CommaText;
     Result := True;
   end else
    Result := False;

  PortNames.Free;

end;

procedure TSerial.SetPort(const s : string);
begin

  FSer.Active := False;
  FPort := s;
  if FPort = NO_PORT_STR then Exit;

  try

    FSer.Device := FPort;
    FSer.Active := True;
    SetBaudrate(FBaudrate);

  except
    raise Exception.Create(ErrMsgSetPort + ' ("' + FPort + '")');
  end;

end;

constructor TSerial.Create;
begin

  inherited;
  FSer := TSdpoSerial.Create(nil);
  GetAvailablePorts;
  // If there's only one port available, connect to it
  if AvailablePorts.Count = 1 then
    try SetPort(AvailablePorts.Strings[0]); except FPort := ''; end
   else FPort := '';
  SetBaudrate(StrToInt(DEFAULT_BAUDRATE_STR));

end;

procedure TSerial.ResetDTR;
begin

  try
    FSer.SetDTR(False);
    Sleep(DTRResetLengthMS);
    FSer.SetDTR(True);

  except
    raise Exception.Create(ErrMsgResetDTR);
  end;

end;

procedure TSerial.Send(const s : string);
begin
  FSer.WriteData(s);
end;

procedure TSerial.AssignRxHandler(p : TNotifyEvent);
begin
  FSer.OnRxData := p;
end;

function TSerial.Receive : string;
begin
  Result := FSer.ReadData;
end;

initialization

  try
    RegAccess := TRegistry.Create;
    RegAccess.RootKey := HKEY_LOCAL_MACHINE;
  except
    raise Exception.Create(ErrMsgRegistryAccess);
  end;

  Serial := TSerial.Create;


finalization

  Serial.Free;
  if assigned(RegAccess) then RegAccess.Free;

end.

