unit SMDUtils;

{$mode objfpc}{$H+}

interface
uses
  Classes, SysUtils;

resourcestring
  ErrSaveToFile = 'Error: Saving to file failed';
  ErrLoadFromFile = 'Error: Loading from file failed';

const
  Escape : ANSIChar = '\';

  HISTORY_MAX_LINES = 32;
  HISTORY_FILENAME  = 'SendHistory.txt';

  BAUD_ABSOLUTE_MIN        = 110;
  BAUD_ABSOLUTE_MAX        = 2500000;
  DEFAULT_BAUDRATE_STR     = '9600';
  ARDUINO_BAUDRATES_STR    = '300,1200,2400,4800,9600,19200,57600,115200';
  CUSTOM_BAUDRATE_FILENAME = 'CustomBaudrates.txt';

  ENVELOPE_FILENAME   = 'Envelopes.txt';
  AS_IS_ENVELOPE_NAME = 'As is';

type
  TSendHistory = class

    private
      FLines : TStringList;
      function GetCount : Integer;

    public
      constructor Create; overload;
      procedure Add(const s : string);
      procedure Clear;
      procedure SaveToFile;
      procedure CopyTo(sl : TStrings);
      destructor Destroy; override;

      property Count : Integer read GetCount;

  end;

 TBaudManager = class

   private
     FStdBauds, FCustBauds : TStringList;
     function GetStdRatesCount : Integer;
     function GetCustomAsCommaText : string;

   public
     constructor Create; overload;
     procedure Add(BaudStr : string);
     procedure CopyTo(sl : TStrings);
     procedure CopyCustomFrom(sl : TStringList);
     function LoadFromFile(const FileName : string) : Boolean;
     destructor Destroy; override;

     property StdBaudratesCount : Integer read GetStdRatesCount;
     property CustomBaudsAsCommaText : string read GetCustomAsCommaText;

 end;

PEnvelope = ^TEnvelope;
TEnvelope = record
    Name, Before, After : string;
  end;

 TEnvelopeManager = class
   private
     FEnvelopes : TList;
     procedure Erase;
   public
     constructor Create; overload;
     function Find(N : string) : PEnvelope;
     procedure Add(N, B, A : string);
     procedure Delete(N : string);
     procedure CopyTo(sl : TStrings; const IncludeDefault : Boolean = True);
     procedure LoadFromFile;
     procedure SaveToFile;
     destructor Destroy; override;

 end;

function EscapedStrToRaw(Src : string; var Dest : string;
                         const EmptyIsValid : Boolean = False) : Boolean;
function StrToEscaped(s : string) : string;
function IsValidBaudrateList(sl : TStringList) : Boolean;

var
  SendHistory : TSendHistory;
  BaudManager : TBaudManager;
  EnvelopeManager : TEnvelopeManager;
  StartupBaudStr : string = DEFAULT_BAUDRATE_STR;

implementation
uses
  StrUtils;

function IsValidBaudrateList(sl : TStringList) : Boolean;
var
  i, p : Integer;

begin

  Result := False;

  for i := 0 to sl.Count - 1 do
    try

      if (Length(sl[i]) > 0) AND (sl[i][1] = ESCAPE) then
       begin
         sl[i] := Copy(sl[i], 2, Length(sl[i]) - 1);
         StartupBaudStr := sl[i];
       end;

      p := StrToInt(sl[i]);
      if (p < BAUD_ABSOLUTE_MIN) OR (p > BAUD_ABSOLUTE_MAX) then
       raise Exception.Create('Inner: Invalid baudrate in list');

    except
      Exit;
    end;

  Result := True;

end;

function TSendHistory.GetCount : Integer;
begin
  Result := FLines.Count;
end;

constructor TSendHistory.Create;
begin
  FLines := TStringList.Create;
  if FileExists(HISTORY_FILENAME) then FLines.LoadFromFile(HISTORY_FILENAME);
end;

procedure TSendHistory.Add(const s : string);
var
  p : Integer;

begin

  // If string already exists in history, delete it
  p := FLines.IndexOf(s);
  if p >= 0 then FLines.Delete(p);
  // Put string at top of history
  FLines.Insert(0, s);
  // Truncate history if too long
  if FLines.Count > HISTORY_MAX_LINES then FLines.Delete(HISTORY_MAX_LINES);
  SaveToFile;

end;

procedure TSendHistory.Clear;
begin
  FLines.Clear;
end;

procedure TSendHistory.CopyTo(sl : TStrings);
begin
  sl.Assign(FLines);
end;

procedure TSendHistory.SaveToFile;
begin

  try
    FLines.SaveToFile(HISTORY_FILENAME);
  except
    raise Exception.Create(ErrSaveToFile + ' ("' + HISTORY_FILENAME + '")');
  end;

end;

destructor TSendHistory.Destroy;
begin
  FLines.Free;
  inherited;
end;

// Returns True if string is non-empty and valid
function EscapedStrToRaw(Src : string; var Dest : string;
                         const EmptyIsValid : Boolean = False) : Boolean;
var
  Offset    : Integer;
  p1, p2, r : Integer;
  Segment   : string;
  b         : Byte;

begin

  Result := False;

  Dest := Src;
  Offset := 1;
  repeat

    p1 := PosEx(Escape, Dest, Offset);
    // No more escaped content?
    if p1 <= 0 then Break;
    p2 := PosEx(Escape, Dest, p1 + 1);
    // Opened without closing?
    if p2 <= 0 then Exit;

    Segment := copy(Dest, p1 + 1, p2 - p1 - 1);
    if Segment = '' then b := Ord(Escape)
     else try

           r := StrToInt(Segment);
           if (r < 0) OR (r > 255) then
            raise Exception.Create('');
           b := r;

          except
            Exit;
          end;

    Delete(Dest, p1, p2 - p1 + 1);
    Insert(Chr(b), Dest, p1);
    Offset := p1 + 1;

  until Offset > Length(Dest);

  Result := (Dest <> '') OR EmptyIsValid;

end;

function StrToEscaped(s : string) : string;
var
  i : Integer;

begin

  Result := '';
  for i := 1 to Length(s) do
   if s[i] in [#10, #13, #32..#127] then Result := Result + s[i]
    else Result := Result + Char(Escape) + IntToStr(Ord(s[i])) + Char(Escape);

end;

function TBaudManager.GetStdRatesCount : Integer;
begin
  Result := FStdBauds.Count;
end;

function TBaudManager.GetCustomAsCommaText : string;
begin
  Result := FCustBauds.CommaText;
end;

constructor TBaudManager.Create;
begin

  inherited;

  FStdBauds := TStringList.Create;
  FStdBauds.CommaText := ARDUINO_BAUDRATES_STR;

  FCustBauds := TStringList.Create;
  LoadFromFile(CUSTOM_BAUDRATE_FILENAME);

end;

// Assumes the input is a valid CommaText
procedure TBaudManager.Add(BaudStr : string);
var
  newBauds : TStringList;
  i        : Integer;

begin

  NewBauds := TStringList.Create;
  NewBauds.CommaText := BaudStr;

  for i := 0 to NewBauds.Count - 1 do
   if FStdBauds.IndexOf(NewBauds[i]) < 0 then FCustBauds.Add(NewBauds[i]);

  NewBauds.Free;

end;

procedure TBaudManager.CopyTo(sl : TStrings);
var
  i : Integer;

begin
  sl.Assign(FStdBauds);
  for i := 0 to FCustBauds.Count - 1 do sl.Add(FCustBauds[i]);
end;

procedure TBaudManager.CopyCustomFrom(sl : TStringList);
var
  i : Integer;

begin

  FCustBauds.Clear;
  // Add excluding duplicates
  for i := 0 to sl.Count - 1 do
   if (FStdBauds.IndexOf(sl[i]) < 0) AND (FCustBauds.IndexOf(sl[i]) < 0) then
    FCustBauds.Add(sl[i]);

end;

function TBaudManager.LoadFromFile(const FileName : string) : Boolean;
var
  Data   : TStringList;

begin

  Result := False;
  if NOT FileExists(FileName) then Exit;
  Data := TStringList.Create;
  Data.LoadFromFile(FileName);
  if IsValidBaudrateList(Data) then
   begin
     CopyCustomFrom(Data);
     Result := True;
   end;
  Data.Free;

end;

destructor TBaudManager.Destroy;
begin

  FCustBauds.Add(ESCAPE + StartupBaudStr);
  FCustBauds.SaveToFile(CUSTOM_BAUDRATE_FILENAME);
  FCustBauds.Free;
  FStdBauds.Free;
  inherited;

end;

procedure TEnvelopeManager.Erase;
begin
  while FEnvelopes.Count > 0 do
   Delete(PEnvelope(FEnvelopes.Items[0])^.Name);
end;

constructor TEnvelopeManager.Create;
begin

  inherited;
  FEnvelopes := TList.Create;
  LoadFromFile;

end;

function TEnvelopeManager.Find(N : string) : PEnvelope;
var
  i : Integer;

begin

  N := Upcase(N);
  for i := 0 to FEnvelopes.Count - 1 do
   begin
     Result := PEnvelope(FEnvelopes[i]);
     if Upcase(Result^.Name) = N then Exit;
   end;

  Result := nil;

end;

procedure TEnvelopeManager.Add(N, B, A : string);
var
  p : PEnvelope;

begin

  p := Find(N);
  if p = nil then
   begin
     p := new(PEnvelope);
     FEnvelopes.Add(p);
   end;

  p^.Name := N;
  p^.Before := B;
  p^.After := A;

end;

procedure TEnvelopeManager.Delete(N : string);
var
  i : Integer;
  p : PEnvelope;

begin

  N := Upcase(N);
  for i := 0 to FEnvelopes.Count - 1 do
   begin

     p := PEnvelope(FEnvelopes[i]);
     if Upcase(p^.Name) = N then
      begin

        FEnvelopes.Delete(i);
        Dispose(p);
        Break;

      end;

   end;

end;

procedure TEnvelopeManager.CopyTo(sl : TStrings; const IncludeDefault : Boolean = True);
var
  i : Integer;

begin

  sl.Clear;
  if IncludeDefault then i := 0 else i := 1;
  for i := i to FEnvelopes.Count - 1 do
   sl.Add(PEnvelope(FEnvelopes[i])^.Name);

end;

procedure TEnvelopeManager.LoadFromFile;
var
  f       : TextFile;
  N, B, A : string;

begin

  Erase;
  Add(AS_IS_ENVELOPE_NAME, '', '');
  if NOT FileExists(ENVELOPE_FILENAME) then Exit;

  AssignFile(f, ENVELOPE_FILENAME);
  try

    Reset(f);
    while NOT Eof(f) do
     begin

       Readln(f, N);
       Readln(f, B);
       Readln(f, A);
       Add(N, B, A);

     end;
    CloseFile(f);

  except
    CloseFile(f);
    raise Exception.Create(ErrLoadFromFile + ' ("' + ENVELOPE_FILENAME + '")');
  end;

end;

procedure TEnvelopeManager.SaveToFile;
var
  f : TextFile;
  i : Integer;
  p : PEnvelope;

begin

  AssignFile(f, ENVELOPE_FILENAME);
  try

    Rewrite(f);

    // Do NOT save the default - that's hard-coded
    for i := 1 to FEnvelopes.Count - 1 do
     begin

       p := PEnvelope(FEnvelopes.Items[i]);
       Writeln(f, p^.Name);
       Writeln(f, p^.Before);
       Writeln(f, p^.After);

     end;

    CloseFile(f);

  except
    raise Exception.Create(ErrSaveToFile + ' ("' + ENVELOPE_FILENAME + '")');
  end;

end;

destructor TEnvelopeManager.Destroy;
begin

  Erase;
  FEnvelopes.Free;
  inherited;

end;

initialization

  SendHistory     := TSendHistory.Create;
  BaudManager     := TBaudManager.Create;
  EnvelopeManager := TEnvelopeManager.Create;

finalization

  EnvelopeManager.Free;
  BaudManager.Free;
  SendHistory.SaveToFile;
  SendHistory.Free;

end.

